"""QA Dashboard."""
import logging
import os

import numpy as np
import pandas as pd

from ....garjus import Garjus
from ..utils import file_age


logger = logging.getLogger('dashboard.qa.data')


# TODO: modify save and filter so we save the data before filtering,
# then we don't have to refresh or really do anything, either filter is on or
# off. problem is we are filtering before merging scans/assessors so
# need to refactor that. for now it will be 2 clicks to change to autofilter
# including refresh click.

SCAN_STATUS_MAP = {
    'usable': 'P',
    'questionable': 'P',
    'unusable': 'F'}


ASSR_STATUS_MAP = {
    'Passed': 'P',
    'Good': 'P',
    'Passed with edits': 'P',
    'Questionable': 'P',
    'Failed': 'F',
    'Bad': 'F',
    'Needs QA': 'Q',
    'Do Not Run': 'N'}


QA_COLS = [
    'SESSION', 'SUBJECT', 'PROJECT', 'SCANID', 'ASSR',
    'SITE', 'NOTE', 'DATE', 'TYPE', 'STATUS',
    'ARTTYPE', 'SCANTYPE', 'PROCTYPE', 'XSITYPE', 'SESSTYPE', 'MODALITY',
    'FRAMES', 'DURATION', 'TR', 'THICK', 'SENSE', 'MB', 'RESOURCES',
    'JOBDATE', 'TIMEUSED', 'MEMUSED'
]


def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/qadata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def run_refresh(projects):
    filename = get_filename()

    # force a requery
    df = get_data(projects)

    save_data(df, filename)

    return df


def update_data(projects):
    fname = get_filename()

    # Load what we have now
    df = read_data(fname)

    # Remove projects not selected
    df = df[df.PROJECT.isin(projects)]

    # Find new projects in selected
    new_projects = [x for x in projects if x not in df.PROJECT.unique()]

    if new_projects:

        # Save a file with new projects placeholders (hacky lock)
        for p in new_projects:
            _newdf = pd.DataFrame.from_records([{'PROJECT': p}])
            df = pd.concat([df, _newdf], ignore_index=True)

        save_data(df, fname)

        # Load the new projects
        dfp = get_data(new_projects)

        # Merge our new data with old data
        df = read_data(fname)
        df = df[~df.PROJECT.isin(new_projects)]
        df = pd.concat([df, dfp])

        # Save it to file
        save_data(df, fname)

    return df


def load_data(projects=[], refresh=False, maxmins=60, hidetypes=True):
    demodir = os.path.expanduser("~/.garjus/DashboardDemoUser/DATA")
    if os.path.exists(demodir):
        # We are in a demo
        fname = f'{demodir}/qadata.pkl'
        logger.info(f'reading demo data:{fname}')
        df = read_data(fname)
    else:
        fname = get_filename()

        if not os.path.exists(fname):
            refresh = True
        elif file_age(fname) > maxmins:
            logger.info(f'refreshing, file age limit reached:{maxmins} minutes')
            refresh = True

        if refresh:
            df = run_refresh(projects)
        elif set(projects) != set(read_data(fname).PROJECT.unique()):
            logger.debug('updating data')
            # Different projects selected, update
            df = update_data(projects)
        else:
            df = read_data(fname)

        if df.empty:
            return df

        if hidetypes:
            logger.debug('applying autofilter to hide unused types')
            scantypes = None
            assrtypes = None

            garjus = Garjus()

            if garjus.redcap_enabled():
                # Load types
                logger.debug('loading scan/assr types')
                scantypes = garjus.all_scantypes()
                assrtypes = garjus.all_proctypes()

                # Make the lists unique
                scantypes = list(set(scantypes))
                assrtypes = list(set(assrtypes))

                if not scantypes and not df.empty:
                    # Get list of scan types based on assessor inputs
                    logger.debug('loading used scan types')
                    scantypes = garjus.used_scantypes(
                        df[df.TYPE == 'ASSR'],
                        df[df.TYPE == 'SCAN']
                    )

                # Apply filter
                alltypes = scantypes + assrtypes

                if alltypes is not None:
                    logger.debug(f'filtering by types:{len(df)}')
                    df = df[df.TYPE.isin(alltypes)]

            logger.debug(f'done filtering by types:{len(df)}')

    # Filter projects
    df = df[df['PROJECT'].isin(projects)]

    # Must have type
    df = df.dropna(subset=['TYPE'])
    df = df[df.TYPE != '']

    return df


def read_data(filename):
    df = pd.read_pickle(filename)

    if df is None or len(df) == 0:
        df = pd.DataFrame(columns=['PROJECT', 'SCANTYPE', 'SESSTYPE', 'PROCTYPE'])

    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def get_data(projects):
    df = pd.DataFrame(columns=QA_COLS)

    if not projects:
        # No projects selected so we don't query
        return df

    try:
        garjus = Garjus()

        # Load data
        logger.debug(f'load data:{projects}')
        logger.debug(f'load scan data:{projects}')
        scan_df = load_scan_data(garjus, projects)
        logger.debug(f'load assr data:{projects}')
        assr_df = load_assr_data(garjus, projects)
        logger.debug(f'load sgp data:{projects}')
        subj_df = load_sgp_data(garjus, projects)

        logger.debug(f'load subjects:{projects}')
        if garjus.redcap_enabled():
            subjects = load_subjects(garjus, projects)
        else:
            subjects = None

        logger.debug(f'all loaded')
    except Exception as err:
        logger.error(f'load failed:{err}')
        _cols = QA_COLS + ['DATE', 'SESSIONLINK', 'SUBJECTLINK']
        return pd.DataFrame(columns=_cols)

    logger.debug(f'merging data:{projects}')

    # Make a common column for type
    assr_df['TYPE'] = assr_df['PROCTYPE']
    scan_df['TYPE'] = scan_df['SCANTYPE']
    subj_df['TYPE'] = subj_df['PROCTYPE']
    assr_df['ARTTYPE'] = 'assessor'
    scan_df['ARTTYPE'] = 'scan'
    subj_df['ARTTYPE'] = 'sgp'

    for x in ['SESSION', 'SITE', 'NOTE', 'SESSTYPE', 'MODALITY']:
        subj_df[x] = 'SGP'

    for x in ['SCANID', 'SCANTYPE', 'FRAMES', 'DURATION', 'TR', 'THICK', 'SENSE', 'MB']:
        assr_df[x] = None
        subj_df[x] = None

    for x in ['JOBDATE', 'TIMEUSED', 'MEMUSED']:
        scan_df[x] = None
        subj_df[x] = None

    assr_df['RESOURCES'] = ''
    subj_df['RESOURCES'] = ''

    for x in ['PROCTYPE', 'ASSR']:
        scan_df[x] = None

    # Concatenate the common cols to a new dataframe
    df = pd.concat([assr_df[QA_COLS], scan_df[QA_COLS]], sort=False)
    df = pd.concat([df[QA_COLS], subj_df[QA_COLS]], sort=False)

    df['DATE'] = df['DATE'].dt.strftime('%Y-%m-%d')

    if subjects is None:
        df['GROUP'] = 'UNKNOWN'
        df['AGE'] = ''
        df['SEX'] = ''
    else:
        df = pd.merge(
            df,
            subjects,
            left_on=('SUBJECT', 'PROJECT'),
            right_on=('ID', 'PROJECT'),
            how='left'
        )  

    # Convert duration from string of total seconds to formatted HH:MM:SS
    df['DURATION'] = df['DURATION'].fillna(np.nan).replace(
        '', np.nan).replace('None', np.nan)
    df['DURATION'] = pd.to_datetime(
        df.DURATION.astype(float),
        unit='s',
        errors='coerce').dt.strftime("%-M:%S")

    df['SESSIONLINK'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION']

    df['SUBJECTLINK'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT']

    df['PDF'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION'] + \
        '/assessors/' + df['ASSR'] + \
        '/out/resources/PDF/files/' + \
        'report_' + df['ASSR'] + '.pdf'

    df['LOG'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION'] + \
        '/assessors/' + df['ASSR'] + \
        '/out/resources/OUTLOG/files/' + \
        df['ASSR'] + '.txt'

    df['NIFTI'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION'] + \
        '/scans/' + df['SCANID'] + \
        '/resources/NIFTI/files?format=zip'

    df['JSON'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION'] + \
        '/scans/' + df['SCANID'] + \
        '/resources/JSON/files?format=zip'

    df['EDAT'] = garjus.xnat_host() + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION'] + \
        '/scans/' + df['SCANID'] + \
        '/resources/EDAT/files?format=zip'

    df.loc[df.RESOURCES.str.contains('EDAT') == False, 'EDAT'] = ''
    df.loc[df.RESOURCES.str.contains('JSON') == False, 'JSON'] = ''
    df.loc[df.RESOURCES.str.contains('NIFTI') == False, 'NIFTI'] = ''

    return df


def _filter(scan_df, assr_df, scantypes, assrtypes):

    # Apply filters
    if scantypes is not None:
        logger.debug(f'filtering scan by types:{len(scan_df)}')
        scan_df = scan_df[scan_df['SCANTYPE'].isin(scantypes)]

    if assrtypes is not None:
        logger.debug(f'filtering assr by types:{len(assr_df)}')
        assr_df = assr_df[assr_df['PROCTYPE'].isin(assrtypes)]

    logger.debug(f'done filtering by types:{len(scan_df)}:{len(assr_df)}')

    return scan_df, assr_df


def load_subjects(garjus, project_filter):
    subjects = pd.DataFrame(
        [], columns=['ID', 'PROJECT', 'GROUP', 'SEX', 'AGE']
    )
    for p in project_filter:
        logger.debug(f'loading subjects:{p}')
        subjects = pd.concat([subjects, garjus.subjects(p).reset_index()])

    return subjects


def load_assr_data(garjus, project_filter):
    dfa = garjus.assessors(project_filter).copy()

    # Drop any rows with empty proctype
    dfa.dropna(subset=['PROCTYPE'], inplace=True)
    dfa = dfa[dfa.PROCTYPE != '']

    # Create shorthand status
    dfa['STATUS'] = dfa['QCSTATUS'].map(ASSR_STATUS_MAP).fillna('Q')

    # Handle failed jobs
    dfa.loc[dfa.PROCSTATUS == 'JOB_FAILED', 'STATUS'] = 'X'

    # Handle running jobs
    dfa.loc[dfa.PROCSTATUS == 'JOB_RUNNING', 'STATUS'] = 'R'

    # Handle NEED INPUTS
    dfa.loc[dfa.PROCSTATUS == 'NEED_INPUTS', 'STATUS'] = 'N'

    return dfa


def load_sgp_data(garjus, project_filter):

    df = garjus.subject_assessors(project_filter).copy()

    # Get subset of columns
    df = df[[
        'PROJECT', 'SUBJECT', 'DATE', 'ASSR', 'QCSTATUS', 'XSITYPE',
        'PROCSTATUS', 'PROCTYPE']]

    df.drop_duplicates(inplace=True)

    # Drop any rows with empty proctype
    df.dropna(subset=['PROCTYPE'], inplace=True)
    df = df[df.PROCTYPE != '']

    # Create shorthand status
    df['STATUS'] = df['QCSTATUS'].map(ASSR_STATUS_MAP).fillna('Q')

    # Handle failed jobs
    df.loc[df.PROCSTATUS == 'JOB_FAILED', 'STATUS'] = 'X'

    # Handle running jobs
    df.loc[df.PROCSTATUS == 'JOB_RUNNING', 'STATUS'] = 'R'

    # Handle NEED INPUTS
    df.loc[df.PROCSTATUS == 'NEED_INPUTS', 'STATUS'] = 'N'

    return df


def load_scan_data(garjus, project_filter):

    #  Load data
    dfs = garjus.scans(project_filter)

    dfs = dfs[[
        'PROJECT', 'SESSION', 'SUBJECT', 'NOTE', 'DATE', 'SITE', 'SCANID',
        'SCANTYPE', 'QUALITY', 'XSITYPE', 'SESSTYPE', 'MODALITY',
        'FRAMES', 'DURATION', 'TR', 'THICK', 'SENSE', 'MB', 'RESOURCES',
        'full_path']].copy()

    dfs.drop_duplicates(inplace=True)

    # Drop any rows with empty type
    dfs.dropna(subset=['SCANTYPE'], inplace=True)

    # Create shorthand status
    dfs['STATUS'] = dfs['QUALITY'].map(SCAN_STATUS_MAP).fillna('U')

    return dfs


def filter_data(df, projects, proctypes, scantypes, starttime, endtime, sesstypes):

    # Filter by project
    if projects:
        logger.debug('filtering by project:')
        logger.debug(projects)
        df = df[df['PROJECT'].isin(projects)]

    # Filter by proc type
    if proctypes:
        logger.debug('filtering by proc types:')
        logger.debug(proctypes)
        df = df[(df['PROCTYPE'].isin(proctypes)) | (df['ARTTYPE'] == 'scan')]

    # Filter by scan type
    if scantypes:
        logger.debug('filtering by scan types:')
        logger.debug(scantypes)
        df = df[(df['SCANTYPE'].isin(scantypes)) | (df['ARTTYPE'] == 'assessor') | (df['ARTTYPE'] == 'sgp')]

    if starttime:
        logger.debug(f'filtering by start time:{starttime}')
        df = df[pd.to_datetime(df.DATE) >= starttime]

    if endtime:
        df = df[pd.to_datetime(df.DATE) <= endtime]

    # Filter by sesstype
    if sesstypes:
        df = df[df['SESSTYPE'].isin(sesstypes)]

    return df



import logging
import os
import pandas as pd
from datetime import datetime

from ....garjus import Garjus


logger = logging.getLogger('dashboard.reports.data')


def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/reportsdata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def run_refresh(filename):
    df = get_data()

    save_data(df, filename)

    return df


def load_options(df):
    projects = []
    types = []
    times = ['All', 'Current']

    # Projects
    garjus = Garjus()
    projects = garjus.projects()

    # Selected types
    types = df.TYPE.unique()

    # Remove blanks and sort
    types = [x for x in types if x]
    types = sorted(types)

    return projects, types, times


def load_data(refresh=False):
    filename = get_filename()

    if refresh or not os.path.exists(filename):
        run_refresh(filename)

    logger.info('reading data from file:{}'.format(filename))
    df = read_data(filename)

    return df


def filter_data(df, projects, types, timeframe):

    if projects:
        df = df[df.PROJECT.isin(projects)]

    if types:
        df = df[df.TYPE.isin(types)]

    if timeframe == 'Current':
        cur_double = datetime.now().strftime("%B%Y")
        df = df[df.NAME == cur_double]

    return df


def read_data(filename):
    df = pd.read_pickle(filename)
    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def get_data():
    g = Garjus()

    if not g.redcap_enabled():
        logger.debug('redcap not enabled, no reports data')
        return pd.DataFrame(columns=g.column_names('reports'))

    # Get the pid of the main redcap so we can make links
    pid = g.redcap_pid()

    # Load
    df = g.reports()

    # Make pdf link
    df['VIEW'] = 'https://redcap.vanderbilt.edu/redcap_v14.0.0/DataEntry/index.php?' + \
    'pid=' + str(pid) + \
    '&page=' + df.TYPE.str.lower() + \
    '&id=' + df['PROJECT'] + \
    '&instance=' + df['ID'].astype(str)

    return df



"""

update will create any missing

"""
import logging
import os


logger = logging.getLogger(__name__)


def update(garjus, projects, scantypes=None):
    """Update project progress."""

    if not scantypes:
        scantypes = garjus.all_scantypes()

    for p in projects:
        logger.debug(f'updating project:{p}')
        update_project(garjus, p, scantypes)


def update_project(garjus, project, scantypes):
    """Update project scans."""

    logger.debug(f'loading existing:{project}')
    try:
        # Get list of scans already done
        _scans = garjus.stats_scans(project, scantypes)
        existing = [(x['scan_session'], x['scan_id']) for x in _scans]
    except Exception as err:
        logger.debug(f'cannot load, check key:{project}:{err}')
        return

    # Get scans for selected project and scantypes
    df = garjus.scans([project], scantypes)
    logger.debug(f'total scans:{len(df)}')

    # Filter to remove already uploaded
    df['SESSION_SCANID'] = list(zip(df.SESSION.astype(str), df.SCANID.astype(str)))
    df = df[~df.SESSION_SCANID.isin(existing)]
    logger.debug(f'scans after filtering out already uploaded:{len(df)}')

    # Filter out unusable
    df = df[df['QUALITY'] != 'unusable']
    logger.debug(f'scans after filtering out unusable:{len(df)}')

    # Iterate xnat scans
    for r in df.sort_values('SESSION_SCANID').to_dict('records'):
        try: 
            if 'JSON' in r['RESOURCES']:
                update_scan(
                    garjus,
                    r['PROJECT'],
                    r['SUBJECT'],
                    r['SESSION'],
                    r['SCANID'],
                )
            else:
                logger.debug(f'no JSON resource')

        except ConnectionError as err:
            logger.info(err)
            logger.info('waiting a minute')
            os.sleep(60)


def update_scan(garjus, proj, subj, sess, scan):
    """Update scan stats."""

    try:
        stats = garjus.get_scan_stats(proj, subj, sess, scan)
    except Exception as err:
        logger.warn(f'could not set stats:{sess}:{scan}:{err}')
        return

    if stats.get('duration', False) or stats.get('tr', False) or stats.get('tracer', False):
        # we go something so upload it
        try:
            logger.debug(f'uploading:{proj}:{subj}:{sess}:{scan}')
            garjus.set_scan_stats(proj, subj, sess, scan, stats)
        except Exception as err:
            logger.warn(f'could not set stats:{sess}:{scan}:{err}')
            return
    else:
        logger.debug('nothing to upload')



import logging
import os

from garjus.compare import make_double_report
from garjus import Garjus


logger = logging.getLogger('test_compare_arousal')

PROJECT = 'REMBRANDT'

FIELDS = [
    'arousal_pre_1calm',
    'arousal_pre_2control',
    'arousal_pre_3happy',
    'arousal_post_1calm',
    'arousal_post_2control',
    'arousal_post_3happy',
    'msit_dia1',
    'msit_dia2',
    'msit_dia3',
    'msit_dia4',
    'msit_pulse1',
    'msit_pulse2',
    'msit_pulse3',
    'msit_pulse4',
    'msit_start',
    'msit_stop',
    'msit_sys1',
    'msit_sys2',
    'msit_sys3',
    'msit_sys4',
    't1_dia1',
    't1_dia2',
    't1_dia3',
    't1_pulse1',
    't1_pulse2',
    't1_pulse3',
    't1_sys1',
    't1_sys2',
    't1_sys3',
]

EVENTS = [
    'baselinemonth_0_arm_2',
    'baselinemonth_0_arm_3',
    'month_8_arm_2',
    'month_8_arm_3',
    'month_16_arm_2',
    'month_16_arm_3',
    'month_24_arm_2',
    'month_24_arm_3',
]


def compare(g, project, fields, events):
    """Create a PDF report of Double Entry Comparison."""
    pdf_file = f'{project}_double.pdf'
    excel_file = f'{project}_double.xlsx'

    if os.path.exists(pdf_file):
        logger.info(f'{pdf_file} exists, delete or rename.')
        return

    if os.path.exists(excel_file):
        logger.info(f'{excel_file} exists, delete or rename.')
        return

    logger.info(f'writing report to file:{pdf_file},{excel_file}.')
    # Get the projects to compare
    proj_primary = g.primary(project)
    proj_secondary = g.secondary(project)
    make_double_report(
        proj_primary,
        proj_secondary,
        pdf_file,
        excel_file,
        fields,
        events)


if __name__ == "__main__":
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    logging.info('connecting to redcap')

    g = Garjus()

    compare(g, PROJECT, FIELDS, EVENTS)

    logging.info('Done!')



import pathlib
import tempfile
import subprocess as sb
import os
import io
import logging
import requests
from zipfile import ZipFile

# Inputs: source to be uploaded, either a path to zip, directory, gtudy URL, or
# XNAT urlproject/subject/session

# steps:
# -unzip if needed
# -run dcm2niix to resort into one folder per series/scan
# -run dcm2niix on each series/scan
# -zip the scan (series) folder
# -upload the zip as DICOMZIP
# -upload the NIFTI, JSON, BVAL/BVEC
# -set date, tracer, scan type, description


logger = logging.getLogger('garjus.import_dicom')


def import_dicom(garjus, src, dst):
    logger.debug(f'uploading from:{src}')

    (proj, subj, sess) = dst.split('/')
    logger.debug(f'uploading to:{proj},{subj},{sess}')

    if src.endswith('.zip'):
        import_dicom_zip(garjus, src, proj, subj, sess)
    elif src.startswith('http'):
        # e.g. gstudy link
        import_dicom_url(garjus, src, proj, subj, sess)
    elif os.path.isdir(src):
        import_dicom_dir(garjus, src, proj, subj, sess)
    else:
        logger.error(f'unsupported source specified:{src}')
        return

    logger.info('Please Note! only DICOM that successfullly converts\
        to NIFTI is uploaded as DICOMZIP')


def import_dicom_zip(garjus, zip_path, project, subject, session):
    """Import zip of directory of DICOM files, sorted or not."""
    with tempfile.TemporaryDirectory() as temp_dir:
        unzipped_dir = pathlib.Path(f'{temp_dir}/UNZIPPED')
        unzipped_dir.mkdir()
        dicom_dir = pathlib.Path(f'{temp_dir}/DICOM')
        dicom_dir.mkdir()

        # Unzip the zip to the temp folder
        logger.info(f'unzip {zip_path} to {unzipped_dir}')
        sb.run(['unzip', '-q', zip_path, '-d', unzipped_dir])

        # Rename/sort dicom
        logger.info(f'rename/sort dicom from {unzipped_dir} to {dicom_dir}')
        garjus.rename_dicom(unzipped_dir, dicom_dir)

        # Upload
        logger.info(f'uploading:{dicom_dir}')
        garjus.upload_session(dicom_dir, project, subject, session)


def import_dicom_dir(garjus, dir_path, project, subject, session):
    """Import directory of DICOM files, sorted or not."""
    with tempfile.TemporaryDirectory() as temp_dir:
        dicom_dir = pathlib.Path(f'{temp_dir}/DICOM')
        dicom_dir.mkdir()

        # Rename/sort dicom
        logger.info(f'rename/sort dicom from {dir_path} to {dicom_dir}')
        garjus.rename_dicom(dir_path, dicom_dir)

        # Upload
        logger.info(f'uploading:{dicom_dir}')
        garjus.upload_session(dicom_dir, project, subject, session)


def import_dicom_url(garjus, url_path, project, subject, session):
    """Import url of DICOM files, sorted or not."""
    with tempfile.TemporaryDirectory() as temp_dir:

        dicom_dir = pathlib.Path(f'{temp_dir}/DICOM')
        dicom_dir.mkdir()
        unzipped_dir = pathlib.Path(f'{temp_dir}/UNZIPPED')
        unzipped_dir.mkdir()

        r = requests.get(url_path, verify=False)
        z = ZipFile(io.BytesIO(r.content))
        z.extractall(unzipped_dir)

        # Rename/sort dicom
        logger.info(f'rename/sort dicom from {unzipped_dir} to {dicom_dir}')
        garjus.rename_dicom(unzipped_dir, dicom_dir)

        # Upload
        logger.info(f'uploading:{dicom_dir}')
        garjus.upload_session(dicom_dir, project, subject, session)



"""Analyses."""
import logging
import os
import re
import shutil
import tempfile
import subprocess as sb
import yaml
import zipfile
from datetime import datetime
import zipfile
import glob

import pandas as pd
import requests


logger = logging.getLogger('garjus.analyses')


# Files are downloaded into hierarchy,
# for session assessors:
# /INPUTS/<SUBJECT>/<SESSION>/assessors/<ASSESSOR>/<FILES>
# for subject assessors:
# /INPUTS/<SUBJECT>/assessors/<ASSESSOR>/<FILES>


class Analysis(object):
    def __init__(self, project, subjects, repo, csvfile=None, yamlfile=None):
        self._project = project
        self._subjects = subjects
        self._csvfile = csvfile
        self._repo = repo

        if yamlfile:
            self._yamlfile = yamlfile
            self._processor = self.load_yaml()
        else:
            self._processor = self.load_processor()

    def load_yaml(self):
        # Load yaml contents
        yaml_file = self._yamlfile
        logger.info(f'loading yamlfile:{yaml_file}')
        try:
            with open(yaml_file, "r") as f:
                return yaml.load(f, Loader=yaml.FullLoader)
        except yaml.error.YAMLError as err:
            logger.error(f'failed to load yaml:{yaml_file}:{err}')

    def load_processor(self):
        filename = 'processor.yaml'

        if self._repo.endswith(filename):
            self._yamlfile = self._repo
            self._repo = os.path.abspath(os.path.join(self._repo, '..', '..', '..'))
            return self.load_yaml()
        elif self._repo.startswith('/'):
            self._yamlfile = f'{self._repo}/{filename}'
            return self.load_yaml()
        else:
            # Load processor yaml from repo
            base = 'https://raw.githubusercontent.com'

            p = self._repo.replace(':', '/').split('/')
            if len(p) != 2:
                logger.error(f'failed to parse:{self._repo}')
                return None

            user = p[0]
            repo = p[1]

            logger.info(f'loading:{user=}:{repo=}')

            url = f'{base}/{user}/{repo}/main/{filename}'

            logger.info(f'{url=}')

            # Get the file contents
            try:
                r = requests.get(url, allow_redirects=True)
                if r.content == '':
                    raise Exception('error exporting file from github')
            except Exception as err:
                logger.error(f'downloading file:{err}')
                return None

            # Cache these for downloading repo later
            self._repo_user = user
            self._repo_name = repo

            return yaml.safe_load(r.text)

    def download_covars(self, garjus, inputs_dir):
        try:
            df = garjus.subjects(self._project)
            #print('renaming')
            df.index.name = 'id'
            #print(df)
            df.to_csv(f'{inputs_dir}/covariates.csv')
        except Exception as err:
            logger.error(f'downloading:{err}')
            return None

    def download_repo(self, repo_dir):
        #curl -sL $REPO | tar zxvf - -C $REPODIR --strip-components=1
        user = self._repo_user
        repo = self._repo_name
        branch = 'main'
        url = f'https://github.com/{user}/{repo}/archive/{branch}.zip'
        repo_zip = f'{repo_dir}/{branch}.zip'

        logger.info(f'loading:{url=}')

        # Get the file contents
        try:
            r = requests.get(url, allow_redirects=True)
            if r.content == '':
                raise Exception('error exporting file from github')
        except Exception as err:
            logger.error(f'downloading file:{err}')
            return None

        # Save it
        if r.status_code == 200:
            with open(repo_zip, 'wb') as f:
                f.write(r.content)
        else:
            logger.error(f'download failed:{r.text}')

        # Unzip it
        with zipfile.ZipFile(repo_zip, 'r') as z:
            z.extractall(repo_dir)

    def run(self, garjus, jobdir):
        jobdir = os.path.abspath(jobdir)

        inputs_dir = f'{jobdir}/INPUTS'
        outputs_dir = f'{jobdir}/OUTPUTS'

        logger.info(f'creating INPUTS and OUTPUTS in:{jobdir}')
        _make_dirs(inputs_dir)
        _make_dirs(outputs_dir)

        # Download inputs
        logger.info(f'downloading analysis inputs to {inputs_dir}')
        self.download_inputs(garjus, inputs_dir)

        # Get the code
        logger.info(f'repo={self._repo}')
        if os.path.exists(self._repo):
            repo_dir = self._repo
            logger.info(f'using local repo:{repo_dir}')
        else:
            # Download repository
            repo_dir = f'{jobdir}/REPO'
            logger.info('downloading repo')
            self.download_repo(repo_dir)

        # Copy covars
        if self._csvfile and os.path.exists(self._csvfile):
            #print('copy csv')
            shutil.copy(self._csvfile, f'{jobdir}/INPUTS/covariates.csv')
        else:
            # Download covars
            self.download_covars(garjus, inputs_dir)

        # Run all commands
        self.run_commands(jobdir, repo_dir)

    def download_inputs(self, garjus, inputs_dir):
        errors = []
        processor = self._processor
        project = self._project
        subjects = self._subjects

        logger.info('loading project data')
        assessors = garjus.assessors(projects=[project])
        scans = garjus.scans(projects=[project])
        sgp = garjus.subject_assessors(projects=[project])

        sessions = pd.concat([
            _sessions_from_scans(scans),
            _sessions_from_assessors(assessors)
        ])
        sessions = sessions.drop_duplicates()

        if not subjects:
            # Default to all subjects
            subjects = list(sessions.SUBJECT.unique())

        logger.info(f'subjects={subjects}')

        # What to download for each subject?
        subj_spec = processor['inputs']['xnat']['subjects']

        logger.debug(f'subject spec={subj_spec}')

        for subj in subjects:
            logger.debug(f'subject={subj}')

            # Make the Subject download folder
            subj_dir = f'{inputs_dir}/{subj}'
            _make_dirs(subj_dir)

            # Download the subject as specified in subj_spec
            try:
                logger.info(f'_download_subject={subj}')
                _download_subject(
                    garjus,
                    subj_dir,
                    subj_spec,
                    project,
                    subj,
                    sessions,
                    assessors,
                    sgp,
                    scans)
            except Exception as err:
                logger.debug(err)
                errors.append(subj)
                continue

        # report what's missing
        if errors:
            logger.info(f'errors{errors}')
        else:
            logger.info(f'download complete with no errors!')

        logger.debug('done!')

    def run_commands(self, jobdir, repodir=None):
        command_mode = 'docker'
        processor = self._processor

        # Check for docker command
        if not shutil.which('docker'):
            logger.error('docker not found, cannot run containers')
            return

        command = processor.get('command', None)
        if command is None:
            logger.debug('no command found')
            return

        # Run steps
        logger.info('running analysis steps...')

        # Pre command
        precommand = processor.get('pre', None)
        if precommand:
            # Get the container name or path
            container = precommand['container']
            for c in processor['containers']:
                if c['name'] == container:
                    container = c['source']

            extraopts = precommand.get('extraopts', '')
            args = precommand.get('args', '')
            command_type = precommand.get('type', '')

            logger.info(f'running analysis pre-command:{precommand=}')

            _run_command(
                container,
                extraopts,
                args,
                command_mode,
                command_type,
                jobdir,
                repodir
            )

        # And now the main command must run
        container = command['container']
        for c in processor['containers']:
            if c['name'] == container:
                if 'source' in c:
                    container = c['source']
                else:
                    raise Exception('cannot run in this environment.')

        logger.debug(f'command mode is {command_mode}')

        extraopts = command.get('extraopts', '')
        args = command.get('args', '')
        command_type = command.get('type', '')

        logger.info(f'running main command:{command=}')

        _run_command(
            container,
            extraopts,
            args,
            command_mode,
            command_type,
            jobdir,
            repodir
        )

        # Post command
        post = processor.get('post', None)
        if post:
            # Get the container name or path
            container = post['container']
            for c in processor['containers']:
                if c['name'] == container:
                    container = c['source']

            extraopts = post.get('extraopts', '')
            args = post.get('args', '')
            command_type = post.get('type', '')

            logger.info(f'running post command:{post=}')

            _run_command(
                container,
                extraopts,
                args,
                command_mode,
                command_type,
                jobdir,
                repodir
            )


def parse_list(csv_string):
    """
    Split string on commas including any leading/trailing spaces with split
    """
    return re.split(r'\s*,\s*', csv_string)

def _download_zip(xnat, uri, zipfile):
    # Build the uri to download
    _uri = uri + '?format=zip&structure=simplified'

    response = xnat.get(_uri, stream=True)

    if response.status_code != 200:
        raise FileNotFoundError(uri)

    with open(zipfile, 'wb') as f:
        shutil.copyfileobj(response.raw, f)

    return zipfile


def _download_file_stream(xnat, uri, dst):

    response = xnat.get(uri, stream=True)

    logger.debug(f'download response code:{response.status_code}')

    if response.status_code != 200:
        raise FileNotFoundError(uri)

    if dst.endswith('.txt'):
        # Write text as text
        with open(dst, 'w') as f:
            f.write(response.text)
    else:
        # Copy binary file contents
        with open(dst, 'wb') as f:
            shutil.copyfileobj(response.raw, f)

    return dst


def _run_command(
    container,
    extraopts,
    args,
    command_mode,
    command_type,
    jobdir,
    repodir=None
):
    cmd = None

    if not repodir:
        # Mount to top-level of unextracted zipfile
        repodir = glob.glob(f'{jobdir}/REPO/*/')[0]

    # Build the command string
    if command_mode == 'docker':
        if container.startswith('docker://'):
            # Remove docker prefix
            container = container.split('docker://')[1]

        cmd = 'docker'

        if extraopts:
            extraopts = extraopts.replace('-B', '-v')
            logger.info(f'{extraopts=}')

        if command_type == 'singularity_exec':
            cmd += ' run --rm --entrypoint ""'
        else:
            cmd += ' run'

        cmd += f' -v {jobdir}/INPUTS:/INPUTS'
        cmd += f' -v {jobdir}/OUTPUTS:/OUTPUTS'
        cmd += f' -v {repodir}:/REPO'
        cmd += f' {extraopts} {container} {args}'

    if not cmd:
        logger.debug('invalid command')
        return

    # Run it
    logger.info(cmd)
    os.system(cmd)


def run_analysis(
    garjus,
    project,
    subjects,
    repo,
    jobdir,
    csv=None,
    yamlfile=None
):
    # Run it
    logger.info(f'running analysis')
    Analysis(project, subjects, repo, csv, yamlfile).run(
        garjus, jobdir)

    # That is all
    logger.info(f'analysis done!')

def _sessions_from_scans(scans):
    return scans[[
        'PROJECT',
        'SUBJECT',
        'SESSION',
        'SESSTYPE',
        'MODALITY',
        'DATE',
        'SITE'
    ]].drop_duplicates()


def _sessions_from_assessors(assessors):
    return assessors[[
        'PROJECT',
        'SUBJECT',
        'SESSION',
        'SESSTYPE',
        'DATE',
        'SITE'
    ]].drop_duplicates()


def _download_scan_file(garjus, proj, subj, sess, scan, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Connect to the resource on xnat
    r = garjus.xnat().select_scan_resource(proj, subj, sess, scan, res)

    # TODO: apply regex or wildcards in fmatch
    # res_obj.files()[0].label()).get(fpath)
    # res.files().label()

    r.file(fmatch).get(dst)

    return dst


def _download_file(garjus, proj, subj, sess, assr, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Connect to the resource on xnat
    r = garjus.xnat().select_assessor_resource(proj, subj, sess, assr, res)

    # TODO: apply regex or wildcards in fmatch
    # res_obj.files()[0].label()).get(fpath)
    # res.files().label()

    r.file(fmatch).get(dst)

    return dst


def _download_sgp_resource_zip(xnat, project, subject, assessor, resource, outdir):
    reszip = '{}_{}.zip'.format(assessor, resource)
    respath = 'data/projects/{}/subjects/{}/experiments/{}/resources/{}/files'
    respath = respath.format(project, subject, assessor, resource)

    logger.debug(f'download zip:{respath}:{reszip}')

    # Download the resource as a zip file
    _download_zip(xnat, respath, reszip)

    # Unzip the file to output dir
    logger.debug(f'unzip file {reszip} to {outdir}')
    with zipfile.ZipFile(reszip) as z:
        z.extractall(outdir)

    # Delete the zip
    os.remove(reszip)


def _download_sgp_file(garjus, proj, subj, assr, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Download the file
    uri = f'data/projects/{proj}/subjects/{subj}/experiments/{assr}/resources/{res}/files/{fmatch}'
    _download_file_stream(garjus.xnat(), uri, dst)


def _download_sess_file(garjus, proj, subj, sess, assr, res, fmatch, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Download the file
    uri = f'data/projects/{proj}/subjects/{subj}/experiments/{sess}/assessors/{assr}/resources/{res}/files/{fmatch}'
    logger.debug(uri)
    _download_file_stream(garjus.xnat(), uri, dst)


def _download_first_file(garjus, proj, subj, sess, scan, res, dst):
    # Make the folders for this file path
    _make_dirs(os.path.dirname(dst))

    # Get name of the first file
    src = garjus.xnat().select_scan_resource(
        proj, subj, sess, scan, res).files().get()[0]

    # Download the file
    uri = f'data/projects/{proj}/subjects/{subj}/experiments/{sess}/scans/{scan}/resources/{res}/files/{src}'
    logger.debug(uri)
    _download_file_stream(garjus.xnat(), uri, dst)


def download_sgp_resources(garjus, project, download_dir, proctype, resources, files):

    assessors = garjus.subject_assessors(
        projects=[project],
        proctypes=[proctype]
    )

    assessors = assessors[assessors.PROCSTATUS == 'COMPLETE']

    for i, a in assessors.iterrows():
        proj = a.PROJECT
        subj = a.SUBJECT
        assr = a.ASSR
        dst = f'{download_dir}/{assr}'

        for res in resources:
            # check if it exists


            if files:
                # Download files
                for fmatch in files:
                    # Have we already downloaded it?
                    if os.path.exists(dst):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download file:{assr}:{res}:{fmatch}')
                    try:
                        _download_sgp_file(
                            garjus,
                            proj,
                            subj,
                            assr,
                            res,
                            fmatch,
                            f'{dst}/{res}/{fmatch}'
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{assr}:{res}:{fmatch}:{err}')
                        import traceback
                        traceback.print_exc()
                        raise err
            else:
                logger.debug(f'{proj}:{subj}:{assr}:{res}:{dst}')
                _download_sgp_resource_zip(
                    garjus.xnat(),
                    proj,
                    subj,
                    assr,
                    res,
                    dst)


def download_resources(
    garjus,
    project,
    download_dir,
    proctype,
    resources,
    files,
    sesstypes,
    analysis_id=None,
    sessinclude=None
):

    logger.debug(f'loading data:{project}:{proctype}')

    assessors = garjus.assessors(
        projects=[project],
        proctypes=[proctype],
        sesstypes=sesstypes)

    if sessinclude:
        assessors = assessors[assessors.SESSION.isin(sessinclude)]

    if analysis_id:
        # Get list of subjects for specified analysis and apply as filter
        logger.info(f'analysis={analysis_id}')

        # Get the subject list from the analysis
        a = garjus.load_analysis(project, analysis_id)
        _subjects = a['SUBJECTS'].splitlines()
        logger.debug(f'applying subject filter to include:{_subjects}')
        assessors = assessors[assessors.SUBJECT.isin(_subjects)]

    if assessors.empty and not sesstypes:
        logger.info('loading as sgp')
        return download_sgp_resources(
            garjus, project, download_dir, proctype, resources, files)

    assessors = assessors[assessors.PROCSTATUS == 'COMPLETE']

    for i, a in assessors.iterrows():
        proj = a.PROJECT
        subj = a.SUBJECT
        sess = a.SESSION
        assr = a.ASSR
        dst = f'{download_dir}/{assr}'

        for res in resources:
            # check if it exists

            if files:
                # Download files
                for fmatch in files:
                    # Have we already downloaded it?
                    if os.path.exists(dst):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download file:{assr}:{res}:{fmatch}')
                    try:
                        _download_sess_file(
                            garjus,
                            proj,
                            subj,
                            sess,
                            assr,
                            res,
                            fmatch,
                            f'{dst}/{res}/{fmatch}'
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{assr}:{res}:{fmatch}:{err}')
                        import traceback
                        traceback.print_exc()
                        raise err
            else:
                logger.debug(f'{proj}:{subj}:{sess}:{assr}:{res}:{dst}')
                try:
                    _download_resource(garjus, proj, subj, sess, assr, res, dst)
                except Exception as err:
                    logger.info(f'failed to download:{assr}:{res}')
                    continue


def download_scan_resources(
    garjus, project, download_dir, scantype, resources, files, sesstypes, sessinclude=None):
    logger.debug(f'loading data:{project}:{scantype}')
    scans = garjus.scans(
        projects=[project], scantypes=[scantype], sesstypes=sesstypes)

    if sessinclude:
        scans = scans[scans.SESSION.isin(sessinclude)]

    scans = scans[scans.QUALITY != 'unusable']

    for i, s in scans.iterrows():
        proj = s.PROJECT
        subj = s.SUBJECT
        sess = s.SESSION
        scan = s.SCANID
        dst = f'{download_dir}/{proj}/{subj}/{sess}/{scan}'

        for res in resources:

            # check if it exists
            if res not in s.RESOURCES.split(','):
                logger.debug(f'no resource:{proj}:{subj}:{sess}:{scan}:{res}')
                continue

            if files:
                # Download files
                for fmatch in files:
                    # Have we already downloaded it?
                    if os.path.exists(dst):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download file:{scan}:{res}:{fmatch}')
                    try:
                        _download_scan_file(
                            garjus,
                            proj,
                            subj,
                            sess,
                            scan,
                            res,
                            fmatch,
                            f'{dst}/{res}/{fmatch}'
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{sess}:{scan}:{res}:{fmatch}:{err}')
                        import traceback
                        traceback.print_exc()
                        raise err
            else:
                logger.debug(f'downloading:{proj}:{subj}:{sess}:{scan}:{res}:{dst}')
                _download_scan_resource(garjus, proj, subj, sess, scan, res, dst)


def _download_resource(garjus, proj, subj, sess, assr, res, dst):
    # Make the folders for destination path
    logger.debug(f'makedirs:{dst}')
    _make_dirs(dst)

    # Connect to the resource on xnat
    logger.debug(f'connecting to resource:{proj}:{subj}:{sess}:{assr}:{res}')
    r = garjus.xnat().select_assessor_resource(proj, subj, sess, assr, res)

    # Download resource and extract
    logger.debug(f'downloading to:{dst}')
    r.get(dst, extract=True)

    return dst


def _download_scan_resource(garjus, proj, subj, sess, scan, res, dst):
    # Make the folders for destination path
    logger.debug(f'makedirs:{dst}')
    _make_dirs(dst)

    # Connect to the resource on xnat
    logger.debug(f'connecting to resource:{proj}:{subj}:{sess}:{scan}:{res}')
    r = garjus.xnat().select_scan_resource(proj, subj, sess, scan, res)

    # Download resource and extract
    logger.debug(f'downloading to:{dst}')
    r.get(dst, extract=True)

    return dst


def _make_dirs(dirname):
    try:
        os.makedirs(dirname)
    except FileExistsError:
        pass


def _download_subject_assessors(garjus, subj_dir, sgp_spec, proj, subj, sgp):

    sgp = sgp[sgp.SUBJECT == subj]

    for k, a in sgp.iterrows():

        assr = a.ASSR

        for assr_spec in sgp_spec:
            logger.debug(f'assr_spec={assr_spec}')

            assr_types = assr_spec['types'].split(',')

            logger.debug(f'assr_types={assr_types}')

            if a.PROCTYPE not in assr_types:
                logger.debug(f'skip assr, no match={assr}:{a.PROCTYPE}')
                continue

            for res_spec in assr_spec['resources']:

                try:
                    res = res_spec['resource']
                except (KeyError, ValueError) as err:
                    logger.error(f'reading resource:{err}')
                    continue

                if 'fmatch' in res_spec:
                    # Download files
                    for fmatch in res_spec['fmatch'].split(','):
                        # Where shall we save it?
                        dst = f'{subj_dir}/assessors/{assr}/{fmatch}'

                        # Have we already downloaded it?
                        if os.path.exists(dst):
                            logger.debug(f'exists:{dst}')
                            continue

                        # Download it
                        logger.info(f'download file:{assr}:{res}:{fmatch}')
                        try:
                            _download_sgp_file(
                                garjus,
                                proj,
                                subj,
                                assr,
                                res,
                                fmatch,
                                dst
                            )
                        except Exception as err:
                            logger.error(f'{subj}:{assr}:{res}:{fmatch}:{err}')
                            import traceback
                            traceback.print_exc()
                            raise err
                else:
                    # Download whole resource
                    dst = subj_dir

                    # Have we already downloaded it?
                    if os.path.exists(os.path.join(dst, assr, res)):
                        logger.debug(f'exists:{dst}/{assr}/{res}')
                        continue

                    # Download it
                    logger.info(f'download resource:{subj}:{assr}:{res}')
                    try:
                        _download_sgp_resource_zip(
                            garjus.xnat(),
                            proj,
                            subj,
                            assr,
                            res,
                            dst)

                    except Exception as err:
                        logger.error(f'{subj}:{assr}:{res}:{err}')
                        raise err


def _download_subject(
    garjus,
    subj_dir,
    subj_spec,
    proj,
    subj,
    sessions,
    assessors,
    sgp,
    scans
):

    #  subject-level assessors
    sgp_spec = subj_spec.get('assessors', None)
    if sgp_spec:
        logger.info(f'download_sgp={subj_dir}')
        _download_subject_assessors(
            garjus,
            subj_dir,
            sgp_spec,
            proj,
            subj,
            sgp)

    # Download the subjects sessions
    for sess_spec in subj_spec.get('sessions', []):

        if sess_spec.get('select', '') == 'first-mri':
            subj_mris = sessions[(sessions.SUBJECT == subj) & (sessions.MODALITY == 'MR')]
            if len(subj_mris) < 1:
                logger.debug('mri not found')
                return

            sess = subj_mris.SESSION.iloc[0]

            sess_dir = f'{subj_dir}/{sess}'
            logger.info(f'download_session={sess_dir}')
            _download_session(
                garjus,
                sess_dir,
                sess_spec,
                proj,
                subj,
                sess,
                assessors,
                scans)

        elif 'types' in sess_spec:
            sess_types = parse_list(sess_spec['types'])

            for i, s in sessions[sessions.SUBJECT == subj].iterrows():
                sess = s.SESSION

                # Apply session type filter
                if s.SESSTYPE not in sess_types:
                    logger.debug(f'skip session, no match={sess}:{s.SESSTYPE}')
                    continue

                sess_dir = f'{subj_dir}/{sess}'
                logger.info(f'download_session={sess_dir}')
                _download_session(
                    garjus,
                    sess_dir,
                    sess_spec,
                    proj,
                    subj,
                    sess,
                    assessors,
                    scans)
        else:
            for i, s in sessions[sessions.SUBJECT == subj].iterrows():
                sess = s.SESSION

                sess_dir = f'{subj_dir}/{sess}'
                logger.info(f'download_session={sess_dir}')
                _download_session(
                    garjus,
                    sess_dir,
                    sess_spec,
                    proj,
                    subj,
                    sess,
                    assessors,
                    scans)


def _download_scans(
    garjus,
    sess_dir,
    sess_spec,
    proj,
    subj,
    sess,
    scans
):

    # get the scans for this session
    sess_scans = scans[scans.SESSION == sess]

    for k, s in sess_scans.iterrows():
        scan = s.SCANID

        for scan_spec in sess_spec.get('scans', []):
            logger.debug(f'scan_spec={scan_spec}')

            scan_types = scan_spec['types'].split(',')

            logger.debug(f'scan_types={scan_types}')

            if s.SCANTYPE not in scan_types:
                logger.debug(f'skip scan, no match={scan}:{s.SCANTYPE}')
                continue

            # Get list of resources to download from this scan
            resources = scan_spec.get('resources', [])

            # Check for nifti tag
            if 'nifti' in scan_spec:
                # Add a NIFTI resource using value as fdest
                resources.append({
                    'resource': 'NIFTI',
                    'fdest': scan_spec['nifti']
                })

            for res_spec in resources:
                try:
                    res = res_spec['resource']
                except (KeyError, ValueError) as err:
                    logger.error(f'reading resource:{err}')
                    continue

                if 'fdest' in res_spec:
                    fdest = res_spec['fdest']
                    logger.debug(f'setting fdest:{fdest}')
                    dst = f'{os.path.dirname(sess_dir)}/{fdest}'

                    # Have we already downloaded it?
                    if os.path.exists(dst):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download:{sess}:{scan}:{res}')
                    try:
                        _download_first_file(
                            garjus,
                            proj,
                            subj,
                            sess,
                            scan,
                            res,
                            dst)
                    except Exception as err:
                        logger.error(f'{sess}:{scan}:{res}:first:{err}')
                        raise err
                elif 'fmatch' in res_spec:
                    # Download files
                    for fmatch in res_spec['fmatch'].split(','):

                        # Where shall we save it?
                        dst = f'{sess_dir}/{scan}/{fmatch}'

                        # Have we already downloaded it?
                        if os.path.exists(dst):
                            logger.debug(f'exists:{dst}')
                            continue

                        # Download it
                        logger.info(f'download:{sess}:{scan}:{res}:{fmatch}')
                        try:
                            _download_scan_file(
                                garjus,
                                proj,
                                subj,
                                sess,
                                scan,
                                res,
                                fmatch,
                                dst
                            )
                        except Exception as err:
                            logger.error(f'{sess}:{scan}:{res}:{fmatch}:{err}')
                            raise err
                else:
                    # Download whole resource

                    # Where shall we save it?
                    dst = f'{sess_dir}/{scan}'

                    # Have we already downloaded it?
                    if os.path.exists(os.path.join(dst, res)):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download resource:{sess}:{scan}:{res}')
                    try:
                        _download_scan_resource(
                            garjus,
                            proj,
                            subj,
                            sess,
                            scan,
                            res,
                            dst
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{sess}:{scan}:{res}:{err}')
                        raise err


def _download_session(
    garjus,
    sess_dir,
    sess_spec,
    proj,
    subj,
    sess,
    assessors,
    scans
):

    if 'scans' in sess_spec:
        _download_scans(garjus, sess_dir, sess_spec, proj, subj, sess, scans)

    # get the assessors for this session
    sess_assessors = assessors[assessors.SESSION == sess]

    for k, a in sess_assessors.iterrows():
        assr = a.ASSR

        for assr_spec in sess_spec.get('assessors', []):
            logger.debug(f'assr_spec={assr_spec}')

            assr_types = assr_spec['types'].split(',')

            logger.debug(f'assr_types={assr_types}')

            if a.PROCTYPE not in assr_types:
                logger.debug(f'skip assr, no match on type={assr}:{a.PROCTYPE}')
                continue

            for res_spec in assr_spec['resources']:

                try:
                    res = res_spec['resource']
                except (KeyError, ValueError) as err:
                    logger.error(f'reading resource:{err}')
                    continue

                if 'fmatch' in res_spec:
                    # Download files
                    for fmatch in res_spec['fmatch'].split(','):

                        # Where shall we save it?
                        if 'fdest' in res_spec:
                            print(f'{_fdest=}')
                            _fdest = res_spec['fdest']
                            dst = f'{sess_dir}/assessors/{assr}/{_fdest}'
                        else:
                            dst = f'{sess_dir}/assessors/{assr}/{fmatch}'

                        # Have we already downloaded it?
                        if os.path.exists(dst):
                            logger.debug(f'exists:{dst}')
                            continue

                        # Download it
                        logger.info(f'download file:{proj}:{subj}:{sess}:{assr}:{res}:{fmatch}')
                        try:
                            _download_file(
                                garjus,
                                proj,
                                subj,
                                sess,
                                assr,
                                res,
                                fmatch,
                                dst
                            )
                        except Exception as err:
                            logger.error(f'{subj}:{sess}:{assr}:{res}:{fmatch}:{err}')
                            raise err
                else:
                    # Download whole resource

                    # Where shall we save it?
                    dst = f'{sess_dir}/{assr}'

                    # Have we already downloaded it?
                    if os.path.exists(os.path.join(dst, res)):
                        logger.debug(f'exists:{dst}')
                        continue

                    # Download it
                    logger.info(f'download resource:{proj}:{subj}:{sess}:{assr}:{res}')
                    try:
                        _download_resource(
                            garjus,
                            proj,
                            subj,
                            sess,
                            assr,
                            res,
                            dst
                        )
                    except Exception as err:
                        logger.error(f'{subj}:{sess}:{assr}:{res}:{err}')
                        raise err





aparc_columns = [
    'StructName',
    'NumVert',
    'SurfArea',
    'GrayVol',
    'ThickAvg',
    'ThickStd',
    'MeanCurv',
    'GausCurv',
    'FoldInd',
    'CurvInd']


def get_stats(xnat, fullpath):
    stats = {}

    print('downloading lh.aparc')
    xnat_file = xnat.select(fullpath).resource('SUBJ').file('stats/lh.aparc.stats').get()
    df = pd.read_table(
        xnat_file,
        comment='#',
        header=None,
        names=aparc_columns,
        delim_whitespace=True,
        index_col='StructName'
    )

    # Thickness Average
    stats['fs7_caudmidfront_lh_thickavg'] = str(df.loc['caudalmiddlefrontal', 'ThickAvg'])
    stats['fs7_entorhinal_lh_thickavg'] = str(df.loc['entorhinal', 'ThickAvg'])
    stats['fs7_latorbfront_lh_thickavg'] = str(df.loc['lateralorbitofrontal', 'ThickAvg'])
    stats['fs7_medorbfront_lh_thickavg'] = str(df.loc['medialorbitofrontal', 'ThickAvg'])
    stats['fs7_rostmidfront_lh_thickavg'] = str(df.loc['rostralmiddlefrontal', 'ThickAvg'])
    stats['fs7_supfront_lh_thickavg'] = str(df.loc['superiorfrontal', 'ThickAvg'])

    print('downloading rh.aparc')
    xnat_file = xnat.select(fullpath).resource('SUBJ').file('stats/rh.aparc.stats').get()
    df = pd.read_table(
        xnat_file,
        comment='#',
        header=None,
        names=aparc_columns,
        delim_whitespace=True,
        index_col='StructName'
    )

    # Thickness Average
    stats['fs7_caudmidfront_rh_thickavg'] = str(df.loc['caudalmiddlefrontal', 'ThickAvg'])
    stats['fs7_entorhinal_rh_thickavg'] = str(df.loc['entorhinal', 'ThickAvg'])
    stats['fs7_latorbfront_rh_thickavg'] = str(df.loc['lateralorbitofrontal', 'ThickAvg'])
    stats['fs7_medorbfront_rh_thickavg'] = str(df.loc['medialorbitofrontal', 'ThickAvg'])
    stats['fs7_rostmidfront_rh_thickavg'] = str(df.loc['rostralmiddlefrontal', 'ThickAvg'])
    stats['fs7_supfront_rh_thickavg'] = str(df.loc['superiorfrontal', 'ThickAvg'])

    return stats



import pandas as pd
from garjus import Garjus


g = Garjus()

df = g.assessors(projects=['REMBRANDT'])

df = df[df.PROCTYPE == 'FS7_v1']

for i, row in df.iterrows():
    print(i, row['full_path'])

    if i < 1201:
        print('nope')
        continue

    stats = get_stats(g.xnat(), row['full_path'])
    print('set stats')
    g.set_stats(row['PROJECT'], row['SUBJECT'], row['SESSION'], row['ASSR'], stats)

print('DONE!')



import os
import logging

from flask import Flask, request, redirect, session, jsonify, url_for, render_template
from flask_login import login_user, LoginManager, UserMixin, logout_user, current_user
import dash
from dash import html
import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template

import garjus
from ..garjus import Garjus
from .pages import qa
from .pages import activity
from .pages import issues
from .pages import queue
from .pages import stats
from .pages import analyses
from .pages import processors
from .pages import reports
from . import content

# This file serves the same purpose as index.py but wrapped in a flask app
# with user/password authentication. garjus will return this app when
# login option is requested.

logger = logging.getLogger('garjus.dashboard.login')

# Connect to an underlying flask server so we can configure it for auth
templates = os.path.join(os.path.dirname(garjus.__file__), 'dashboard/templates')
server = Flask(__name__, template_folder=templates)



@server.before_request
def check_login():
    # TODO: use dash pages module to return pages based on user access level
    logger.debug('checking login')
    if request.method == 'GET':
        if request.path in ['/login', '/logout']:
            # nothing to check here
            return
        if is_authenticated():
            logger.debug(f'user is authenticated:{current_user.id}')
            return

        # nothing to check so user must log in
        return redirect(url_for('login'))
    else:
        if current_user:
            if request.path == '/login' or is_authenticated():
                return

        logout_user()
        return


def is_authenticated():
    return current_user and current_user.is_authenticated and Garjus.is_authenticated()


@server.route('/login', methods=['POST', 'GET'])
def login(message=""):
    try:
        if request.method == 'POST':
            if request.form:
                hostname = 'https://xnat.vanderbilt.edu/xnat'
                username = request.form['username']
                password = request.form['password']

                try:
                    # Get the xnat alias token
                    from ..garjus import Garjus
                    logger.debug('Garjus login')
                    Garjus.login(hostname, username, password)

                    login_user(User(username, hostname))

                    # What page do we send?
                    if session.get('url', False):
                        # redirect to original target
                        url = session['url']
                        logger.debug(f'redirecting to target url:{url}')
                        session['url'] = None
                        return redirect(url)
                    else:
                        # redirect to home
                        logger.debug('redirecting to home')
                        return redirect('/')
                except Exception as err:
                    logger.debug(f'login failed:{err}')
                    message = 'login failed, try again'
        else:
            if current_user:
                if current_user.is_authenticated:
                    try:
                        logger.debug('redirecting to /')
                        return redirect('/')
                    except Exception as err:
                        logger.debug(f'cannot log in, try again:{err}')
                        message = 'login failed, try again'
    except Exception as err:
        logger.error(f'login error, route to logout:{err}')
        return logout()

    logger.debug('rendering login.html')
    return render_template('login.html', message=message)


@server.route('/logout', methods=['GET'])
def logout():
    if current_user:
        if current_user.is_authenticated:
            logout_user()
    return render_template('login.html', message="you have been logged out")

# Prep the configs for the app
dbc_css = "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"
assets_path = os.path.join(os.path.dirname(garjus.__file__), 'dashboard/assets')
darkmode = True

if darkmode:
    stylesheets = [dbc.themes.DARKLY, dbc_css]
    load_figure_template("darkly")
else:
    stylesheets = [dbc.themes.FLATLY, dbc_css]
    load_figure_template("flatly")

# Build the dash app with the configs
app = dash.Dash(
    __name__,
    server=server,
    external_stylesheets=stylesheets,
    assets_folder=assets_path,
    suppress_callback_exceptions=True,
)

# Set the title to appear on web pages
app.title = 'dashboard'

server.config.update(SECRET_KEY=os.urandom(24))

# Login manager object will be used to login / logout users
login_manager = LoginManager()
login_manager.init_app(server)
login_manager.login_view = "/login"


class User(UserMixin):
    # User data model. It has to have at least self.id as a minimum
    def __init__(self, username, hostname=None):
        self.id = username
        self.hostname = hostname


@login_manager.user_loader
def load_user(username):
    """This function loads the user by user id."""
    logger.debug(f'loading user:{username}')
    return User(username)

# Set the main content
app.layout = content.get_content(include_logout=True)

logger.debug(f'{app.layout}')

if __name__ == "__main__":
    app.run_server(debug=True)



import os
import sys
import tempfile
import json
import html
import pathlib
import logging
from zipfile import ZipFile, ZIP_DEFLATED

import dax


logger = logging.getLogger('garjus.utils_xnat')

# The scan URI is a hacky way to get a row for each resource of a every
# scan including all modalities. Things go awry when we try to add any
# other columns.
SCAN_URI = '/REST/experiments?xsiType=xnat:imagesessiondata\
&columns=\
project,\
xnat:imagesessiondata/sharing/share/project,\
subject_label,\
session_label,\
session_type,\
xnat:imagesessiondata/note,\
xnat:imagesessiondata/date,\
tracer_name,\
xnat:imagesessiondata/acquisition_site,\
xnat:imagesessiondata/label,\
xnat:imagescandata/id,\
xnat:imagescandata/type,\
xnat:imagescandata/quality,\
xnat:imagescandata/frames,\
xnat:imagescandata/file/label'


# The scan URI is a hacky way to get a row for each assessor. We do
# not try to get a row per resource because that query takes too long.
# The column name is: proc:genprocdata/out/file/label
ASSR_URI = '/REST/experiments?xsiType=xnat:imagesessiondata\
&columns=\
project,\
xnat:imagesessiondata/sharing/share/project,\
subject_label,\
session_label,\
session_type,\
xnat:imagesessiondata/acquisition_site,\
xnat:imagesessiondata/note,\
xnat:imagesessiondata/date,\
xnat:imagesessiondata/label,\
proc:genprocdata/label,\
proc:genprocdata/procstatus,\
proc:genprocdata/proctype,\
proc:genprocdata/validation/status,\
proc:genprocdata/validation/date,\
proc:genprocdata/validation/validated_by,\
proc:genprocdata/jobstartdate,\
proc:genprocdata/walltimeused,\
proc:genprocdata/memused,\
last_modified,\
proc:genprocdata/inputs'


SGP_URI = '/REST/subjects?xsiType=xnat:subjectdata\
&columns=\
project,\
label,\
proc:subjgenprocdata/label,\
proc:subjgenprocdata/date,\
proc:subjgenprocdata/procstatus,\
proc:subjgenprocdata/proctype,\
proc:subjgenprocdata/validation/status,\
proc:subjgenprocdata/inputs,\
last_modified'


SCAN_RENAME = {
    'project': 'PROJECT',
    'subject_label': 'SUBJECT',
    'session_label': 'SESSION',
    'session_type': 'SESSTYPE',
    'tracer_name': 'TRACER',
    'xnat:imagesessiondata/note': 'NOTE',
    'xnat:imagesessiondata/date': 'DATE',
    'xnat:imagesessiondata/acquisition_site': 'SITE',
    'xnat:imagescandata/id': 'SCANID',
    'xnat:imagescandata/type': 'SCANTYPE',
    'xnat:imagescandata/quality': 'QUALITY',
    'xsiType': 'XSITYPE',
    'xnat:imagescandata/file/label': 'RESOURCES',
    'xnat:imagescandata/frames': 'FRAMES',
}

ASSR_RENAME = {
    'project': 'PROJECT',
    'subject_label': 'SUBJECT',
    'session_label': 'SESSION',
    'session_type': 'SESSTYPE',
    'xnat:imagesessiondata/note': 'NOTE',
    'xnat:imagesessiondata/date': 'DATE',
    'xnat:imagesessiondata/acquisition_site': 'SITE',
    'proc:genprocdata/label': 'ASSR',
    'proc:genprocdata/procstatus': 'PROCSTATUS',
    'proc:genprocdata/proctype': 'PROCTYPE',
    'proc:genprocdata/jobstartdate': 'JOBDATE',
    'proc:genprocdata/validation/status': 'QCSTATUS',
    'proc:genprocdata/validation/date': 'QCDATE',
    'proc:genprocdata/validation/validated_by': 'QCBY',
    'xsiType': 'XSITYPE',
    'proc:genprocdata/inputs': 'INPUTS',
    'proc:genprocdata/walltimeused': 'TIMEUSED',
    'proc:genprocdata/memused': 'MEMUSED',
}

SGP_RENAME = {
    'project': 'PROJECT',
    'label': 'SUBJECT',
    'proc:subjgenprocdata/date': 'DATE',
    'proc:subjgenprocdata/label': 'ASSR',
    'proc:subjgenprocdata/procstatus': 'PROCSTATUS',
    'proc:subjgenprocdata/proctype': 'PROCTYPE',
    'proc:subjgenprocdata/validation/status': 'QCSTATUS',
    'proc:subjgenprocdata/inputs': 'INPUTS'}

XSI2MOD = {
    'xnat:eegSessionData': 'EEG',
    'xnat:mrSessionData': 'MR',
    'xnat:petSessionData': 'PET'}


def decode_inputs(inputs):
    if inputs:
        inputs = decode_url_json_string(inputs)
        return inputs
    else:
        return {}


def decode_url_json_string(json_string):
    """
    Load a string representing serialised json into
    :param json_string:
    :return:
    """
    strings = json.loads(html.unescape(json_string),
                         object_pairs_hook=parse_json_pairs)
    return strings


def parse_json_pairs(pairs):
    """
    An object hook for the json.loads method. Used in decode_url_json_string.
    :param pairs:
    :return: A dictionary of parsed json
    """
    sink_pairs = []
    for k, v in pairs:
        if isinstance(k, str):
            k = k.encode('utf-8').decode()
        if isinstance(v, str):
            v = v.encode('utf-8').decode()
        sink_pairs.append((k, v))
    return dict(sink_pairs)


MR_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/visit_id',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/investigator/firstname',
    'xnat:experimentData/investigator/lastname',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/operator',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/UID',
    'xnat:mrSessionData/coil',
    'xnat:mrSessionData/fieldStrength',
    'xnat:mrSessionData/marker',
    'xnat:mrSessionData/stabilization'
]

OTHER_DICOM_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/startTime',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]

MR_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/startTime',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model',
    'xnat:mrScanData/parameters/flip',
    'xnat:mrScanData/parameters/orientation',
    'xnat:mrScanData/parameters/tr',
    'xnat:mrScanData/parameters/ti',
    'xnat:mrScanData/parameters/te',
    'xnat:mrScanData/parameters/sequence',
    'xnat:mrScanData/parameters/imageType',
    'xnat:mrScanData/parameters/scanSequence',
    'xnat:mrScanData/parameters/seqVariant',
    'xnat:mrScanData/parameters/scanOptions',
    'xnat:mrScanData/parameters/acqType',
    'xnat:mrScanData/parameters/pixelBandwidth',
    'xnat:mrScanData/parameters/voxelRes/x',
    'xnat:mrScanData/parameters/voxelRes/y',
    'xnat:mrScanData/parameters/voxelRes/z',
    'xnat:mrScanData/parameters/fov/x',
    'xnat:mrScanData/parameters/fov/y',
    'xnat:mrScanData/parameters/matrix/x',
    'xnat:mrScanData/parameters/matrix/y',
    'xnat:mrScanData/parameters/partitions',
    'xnat:mrScanData/fieldStrength',
    'xnat:mrScanData/marker',
    'xnat:mrScanData/stabilization',
    'xnat:mrScanData/coil'
]

SC_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]

PET_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/acquisition_site',
    'xnat:imageSessionData/scanner',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/study_id',
    'tracer_name',
    'tracer_startTime',
    'tracer_dose',
    'tracer_isotope',
    'tracer_half-life',
]

CT_EXP_ATTRS = [
    'xnat:experimentData/date',
    'xnat:experimentData/visit_id',
    'xnat:experimentData/time',
    'xnat:experimentData/note',
    'xnat:experimentData/investigator/firstname',
    'xnat:experimentData/investigator/lastname',
    'xnat:imageSessionData/scanner/manufacturer',
    'xnat:imageSessionData/scanner/model',
    'xnat:imageSessionData/operator',
    'xnat:imageSessionData/dcmAccessionNumber',
    'xnat:imageSessionData/dcmPatientId',
    'xnat:imageSessionData/dcmPatientName',
    'xnat:imageSessionData/session_type',
    'xnat:imageSessionData/modality',
    'xnat:imageSessionData/UID'
]

PET_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model',
    'xnat:imageScanData/startTime',
    'xnat:petScanData/parameters/orientation',
    'xnat:petScanData/parameters/originalFileName',
    'xnat:petScanData/parameters/systemType',
    'xnat:petScanData/parameters/fileType',
    'xnat:petScanData/parameters/transaxialFOV',
    'xnat:petScanData/parameters/acqType',
    'xnat:petScanData/parameters/facility',
    'xnat:petScanData/parameters/numPlanes',
    'xnat:petScanData/parameters/frames/numFrames',
    'xnat:petScanData/parameters/numGates',
    'xnat:petScanData/parameters/planeSeparation',
    'xnat:petScanData/parameters/binSize',
    'xnat:petScanData/parameters/dataType'
]

CT_SCAN_ATTRS = [
    'xnat:imageScanData/type',
    'xnat:imageScanData/UID',
    'xnat:imageScanData/note',
    'xnat:imageScanData/quality',
    'xnat:imageScanData/condition',
    'xnat:imageScanData/series_description',
    'xnat:imageScanData/documentation',
    'xnat:imageScanData/frames',
    'xnat:imageScanData/scanner/manufacturer',
    'xnat:imageScanData/scanner/model'
]


def check_attributes(src_obj, dest_obj, dtype=None):
    '''Check that attributes on dest match those on src'''

    if dtype is None:
        dtype = src_obj.datatype()

    if dtype == 'xnat:mrSessionData':
        attr_list = MR_EXP_ATTRS
    elif dtype == 'xnat:mrScanData':
        attr_list = MR_SCAN_ATTRS
    elif dtype == 'xnat:scScanData':
        attr_list = SC_SCAN_ATTRS
    elif dtype == 'xnat:petScanData':
        attr_list = PET_SCAN_ATTRS
    elif dtype == 'xnat:ctScanData':
        attr_list = CT_SCAN_ATTRS
    elif dtype == 'xnat:otherDicomScanData':
        attr_list = OTHER_DICOM_SCAN_ATTRS
    else:
        logger.warning(f'unknown Type:{dtype}')
        return

    for a in attr_list:
        src_v = src_obj.attrs.get(a)
        src_v = src_v.replace("\\", "|")
        dest_v = dest_obj.attrs.get(a)
        if src_v != dest_v:
            logger.warning('mismatch:{a}:src={src_v}, dst={dest_v}')
            dest_obj.attrs.set(a, src_v)


def copy_attrs(src_obj, dest_obj, attr_list):
    """ Copies list of attributes form source to destination"""
    try:
        src_attrs = src_obj.attrs.mget(attr_list)
    except IndexError:
        logger.warning('failed with full attributes, trying minimal set')
        attr_list = OTHER_DICOM_SCAN_ATTRS
        src_attrs = src_obj.attrs.mget(attr_list)

    src_list = dict(list(zip(attr_list, src_attrs)))

    # NOTE: For some reason need to set te again b/c a bug somewhere sets te
    # to sequence name
    te_key = 'xnat:mrScanData/parameters/te'
    if te_key in src_list:
        src_list[te_key] = src_obj.attrs.get(te_key)

    dest_obj.attrs.mset(src_list)


def copy_attributes(src_obj, dest_obj):
    '''Copy attributes from src to dest'''
    src_type = src_obj.datatype()

    if src_type == 'xnat:mrSessionData':
        copy_attrs(src_obj, dest_obj, MR_EXP_ATTRS)
    elif src_type == 'xnat:petSessionData':
        copy_attrs(src_obj, dest_obj, PET_EXP_ATTRS)
    elif src_type == 'xnat:ctSessionData':
        copy_attrs(src_obj, dest_obj, CT_EXP_ATTRS)
    elif src_type == 'xnat:mrScanData':
        copy_attrs(src_obj, dest_obj, MR_SCAN_ATTRS)
    elif src_type == 'xnat:petScanData':
        copy_attrs(src_obj, dest_obj, PET_SCAN_ATTRS)
    elif src_type == 'xnat:ctScanData':
        copy_attrs(src_obj, dest_obj, CT_SCAN_ATTRS)
    elif src_type == 'xnat:scScanData':
        copy_attrs(src_obj, dest_obj, SC_SCAN_ATTRS)
    elif src_type == 'xnat:otherDicomScanData':
        copy_attrs(src_obj, dest_obj, OTHER_DICOM_SCAN_ATTRS)
    else:
        logger.error(f'cannot copy attributes, unsupported type:{src_type}')


def copy_res_zip(src_r, dest_r):
    '''
    Copy a resource from XNAT source to XNAT destination using local cache
    in between
    '''
    try:
        # Download zip of resource
        logger.debug('downloading resource as zip')
        cache_z = src_r.get(tempfile.mkdtemp(), extract=False)

        # Upload zip of resource
        logger.debug('uploading resource as zip')
        dest_r.put_zip(cache_z, extract=True)

        # Delete local zip
        os.remove(cache_z)

    except IndexError:
        logger.error('copy failed:{}:{}'.format(cache_z, sys.exc_info()[0]))
        raise


def is_empty_resource(_res):
    '''Check if resource contains any files'''
    f_count = 0
    for f_in in _res.files().fetchall('obj'):
        f_count += 1
        break

    return f_count == 0


def copy_session(src, dst):
    logger.debug('uploading session attributes')
    dst.create(experiments=src.datatype())
    copy_attributes(src, dst)

    # Process each scan of session
    for src_scan in src.scans().fetchall('obj'):
        scan_label = src_scan.label()

        logger.debug(f'processing scan:{scan_label}')
        dst_scan = dst.scan(scan_label)
        copy_scan(src_scan, dst_scan)


def _file_count(res):
    return int(str(res).split('(')[1].split(' files')[0])


def copy_scan(src_scan, dst_scan):
    scan_type = src_scan.datatype()
    if scan_type == '':
        scan_type = 'xnat:otherDicomScanData'

    dst_scan.create(scans=scan_type)
    copy_attributes(src_scan, dst_scan)

    # Process each resource of scan
    for src_res in src_scan.resources().fetchall('obj'):
        res_label = src_res.label()

        logger.debug(f'processing resource:{res_label}')

        file_count = _file_count(src_res)
        if res_label == 'DICOM' and file_count > 1000:
            logger.debug(f'too many files, upload as DICOMZIP:{file_count}')
            dst_res = dst_scan.resource('DICOMZIP')
            copy_res_dicomzip(src_res, dst_res)
        else:
            dst_res = dst_scan.resource(res_label)
            copy_res(src_res, dst_res)


def copy_res_dicomzip(src_res, dst_res):
    '''
    Copy a DICOM resource from XNAT source to XNAT destination DICOMZIP/NIFTI
    '''
    try:
        # Download zip of resource
        logger.debug('downloading resource as zip')
        cache_z = src_res.get(tempfile.mkdtemp(), extract=False)

        # Upload zip of resource
        logger.debug('Uploading resource as zip, no extract')
        dst_res.put_zip(cache_z, extract=False)

        # Delete local zip
        os.remove(cache_z)

    except IndexError:
        logger.error(f'failed to copy')
        raise


def copy_res(src_res, dst_res):
    try:
        logger.debug('copying resource as zip:'.format(src_res.label()))
        copy_res_zip(src_res, dst_res)
        return
    except Exception as err:
        try:
            logger.debug('failed, trying again:{}'.format(src_res.label()))
            copy_res_zip(src_res, dst_res)
            return
        except Exception:
            logger.error(f'failed twice to copy resource as zip:{err}')


def copy_xnat_session(src, dst):
    (src_proj, src_subj, src_sess) = src.split('/')
    (dst_proj, dst_subj, dst_sess) = dst.split('/')

    with dax.XnatUtils.get_interface() as xnat:
        src_sess_obj = xnat.select_session(src_proj, src_subj, src_sess)
        if not src_sess_obj.exists():
            logger.info('src session does not exist')
            return

        dst_proj_obj = xnat.select_project(dst_proj)
        if not dst_proj_obj.exists():
            logger.info('dest project does not exist, refusing to create')
            return

        dst_subj_obj = dst_proj_obj.subject(dst_subj)
        if not dst_subj_obj.exists():
            logger.info('destination subject does not exist, creating')
            dst_subj_obj.create()
        else:
            logger.info('destination subject exists', dst_subj)

        dst_sess_obj = dst_subj_obj.experiment(dst_sess)
        if dst_sess_obj.exists():
            logger.info('destination session exists, refusing to overwrite')
            return

        logger.info(f'destination session does not exist, creating:{dst_sess}')
        copy_session(src_sess_obj, dst_sess_obj)


def refresh_dicom_catalog(xnat, proj, subj, sess, scan):
    _uri = '/data/services/refresh/catalog?resource='
    _uri += f'/archive/projects/{proj}/subjects/{subj}/experiments/{sess}/scans/{scan}/resources/DICOM'
    logger.info('refreshing dicom catalog')
    xnat.post(_uri)


def upload_files(inputfiles, resource):
    logger.debug(f'uploading:{inputfiles}')
    return dax.XnatUtils.upload_files_to_obj(inputfiles, resource, remove=True)


def upload_file(inputfile, resource):
    logger.debug(f'uploading:{inputfile}')
    return dax.XnatUtils.upload_file_to_obj(inputfile, resource, remove=True)


def upload_dirzip(inputdir, resource):

    with tempfile.TemporaryDirectory() as tempdir:
        fzip = os.path.join(
            tempdir,
            '{}.zip'.format(pathlib.Path(inputdir).name))

        logger.info(f'create zip:{inputdir}')
        _create_zip(inputdir, fzip)

        logger.info(f'upload:{fzip}')
        dax.XnatUtils.upload_file_to_obj(fzip, resource)


def _create_zip(input_dir, output_zip):
    dir_path = pathlib.Path(input_dir)
    with ZipFile(output_zip, mode="w", compression=ZIP_DEFLATED, compresslevel=9) as archive:

        for file_path in dir_path.rglob("*.dcm"):
            archive.write(file_path, arcname=file_path.relative_to(dir_path))


def get_admin_projects(xnat):
    """Get result of xnat query."""
    uri = '/data/archive/projects?accessible=true'
    logger.debug(uri)
    json_data = json.loads(xnat._exec(uri, 'GET'), strict=False)
    result = json_data['ResultSet']['Result']

    if len(result) == 0:
        return []

    return [x['id'] for x in result]


def get_my_projects(xnat):
    """Get result of xnat query."""
    uri = '/data/archive/projects?accessible=true'
    logger.debug(uri)
    json_data = json.loads(xnat._exec(uri, 'GET'), strict=False)
    result = json_data['ResultSet']['Result']
    roles = ['Owners', 'Members', 'Collaborators']

    if len(result) == 0:
        return []

    user_role_column = None
    # Search for user role column
    for k in result[0].keys():
        if k.startswith('user_role_'):
            user_role_column = k
            break

    if user_role_column is None:
        logger.info(f'could not determine user role column')
        return []

    logger.debug(f'user role column:{user_role_column}')

    return [x['id'] for x in result if x[user_role_column] in roles]


def get_my_favorites(xnat):
    uri = '/data/archive/projects?favorite=True'
    fav_json = json.loads(xnat._exec(uri, 'GET'), strict=False)
    fav = [x['id'] for x in fav_json['ResultSet']['Result']]
    return fav



import os
import logging

import dash
import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template

from . import content

dbc_css = "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"

logger = logging.getLogger('garjus.dashboard.demo')

# TODO: fix this to find path more dynamically
assets_path = os.path.expanduser('~/git/garjus/garjus/dashboard/assets')

stylesheets = [dbc.themes.DARKLY, dbc_css]

load_figure_template("darkly")

# Build the dash app with the configs
app = dash.Dash(
    __name__,
    external_stylesheets=stylesheets,
    assets_folder=assets_path,
    suppress_callback_exceptions=True,
)

# Set the title to appear on web pages
app.title = 'dashboard'

#server = app.server
#server.config.update(SECRET_KEY=os.urandom(24))

# Set the main content
app.layout = content.get_content(demo=True)

logger.debug(f'{app.layout}')


#app.config.suppress_callback_exceptions = True


if __name__ == "__main__":
    app.run_server(debug=True)



"""dashboard home"""
import logging
from dateutil.relativedelta import relativedelta
from datetime import datetime

import pandas as pd

from .. import queue, issues, reports
from ....garjus import Garjus


logger = logging.getLogger('dashboard.hub.data')


def _get_reports_data(refresh=False):
    df = reports.data.load_data(refresh=refresh)
    return df


def _get_automations_data(g=None, refresh=False):
    if g is None:
        g = Garjus()

    df = pd.DataFrame()
    return df


def _get_processing_data(g=None, refresh=False):
    if g is None:
        g = Garjus()

    df = g.processing_protocols()
    return df


def _get_activity_data(g=None, refresh=False):
    startdate = datetime.today() - relativedelta(days=7)
    startdate = startdate.strftime('%Y-%m-%d')

    if g is None:
        g = Garjus()

    if not g.redcap_enabled():
        return pd.DataFrame(columns=g.column_names('activity'))

    logger.info(f'loading activity:startdate={startdate}')
    df = g.activity(startdate=startdate)
    df.reset_index(inplace=True)
    df['ID'] = df.index

    return df


def _get_queue_data(refresh=False):
    return queue.data.load_data(refresh=refresh, hidedone=True)


def _get_issues_data(refresh=False):
    return issues.data.load_data(refresh=refresh)


def _load_options(df):
    return list(df.PROJECT.unique())



import io
import logging
import os
import tempfile

import numpy as np
import pandas as pd
import redcap


logger = logging.getLogger('garjus.automations.ACOBA_nback')


API_URL = 'https://redcap.vanderbilt.edu/api/'
API_KEY = ''

REDCAP_EVENTS = [
    'screening_arm_1',
]

REDCAP_FIELDS = [
    'participant_id',
    'nback_tabfile',
    'nback_e90d1a_complete',
    'zero_back_fa',
]

M_COUNT = 14.0

X_COUNT = 40.0


def download_file(proj, record_id, event_id, field_id, tmpdir):
    filename = os.path.join(tmpdir, record_id+'_'+event_id+'_'+field_id)
    
    logger.debug(f'downloading file:{record_id}:{event_id}:{field_id}')
    
    try:
        cont,hdr = proj.export_file(record=record_id, event=event_id, field=field_id)
        if cont == '':
            raise RedcapError
    except RedcapError:
        logger.error('downloading file')
        return None
    else:
        with open(filename, 'wb') as f:
            f.write(cont)

        return filename


def parse_nback(nback_file):
    data = {}

    logger.info(f'parsing NBACK:{nback_file}')

    # Determine number of lines to skip
    skip_count = 0
    with open(nback_file) as f:
        for line in f:
            if line.startswith('ExperimentName'):
                break
            else:
                skip_count += 1

    # Read the file
    df = pd.io.parsers.read_csv(
        nback_file,
        skiprows=skip_count,
        header=0,
        index_col=False,
        sep='\t',
    )
    
    # Fix column names
    cols = df.columns
    cols = cols.map(lambda x: x.replace('.', '_').replace(']','_').replace('[','_'))
    df.columns = cols
    
    # Get ID fields 
    subj_list = df['Subject'].unique()
    if len(subj_list) > 1:
        return
    
    sess_list = df['Session'].unique()
    if len(sess_list) > 1:
        return
    
    group_list = df['Group'].unique()
    if len(group_list) > 1:
        return
    
    date_list = df['SessionDate'].unique()
    if len(date_list) > 1:
        return
    
    data['Subject'] = subj_list[0]
    data['Session'] = int(sess_list[0])
    data['Group'] = int(group_list[0])
    data['SessionDate'] = date_list[0]
    
    # Calculate counts
    hit_count_0 = len(df[(df['procedure_Block_'] == 'zeroproc')
                         & (df['LetterM_CRESP'] == 'x') & (df['LetterM_RESP'] == 'm')].index)
    hit_count_1 = len(df[(df['procedure_Block_'] == 'oneproc')
                         & (df['OneBack1_CRESP'] == 'x') & (df['OneBack1_RESP'] == 'm')].index)
    hit_count_2 = len(df[(df['procedure_Block_'] == 'twoproc')
                         & (df['TwoBack1_CRESP'] == 'x') & (df['TwoBack1_RESP'] == 'm')].index)
    hit_count_3 = len(df[(df['procedure_Block_'] == 'threeproc')
                         & (df['ThreeBack1_CRESP'] == 'x') & (df['ThreeBack1_RESP'] == 'm')].index)
    fa_count_0  = len(df[(df['procedure_Block_'] == 'zeroproc')
                         & (df['LetterM_CRESP']  == 'm') & (df['LetterM_RESP'] == 'm')].index)
    fa_count_1  = len(df[(df['procedure_Block_'] == 'oneproc')
                         & (df['OneBack1_CRESP'] == 'm') & (df['OneBack1_RESP'] == 'm')].index)
    fa_count_2  = len(df[(df['procedure_Block_'] == 'twoproc')
                         & (df['TwoBack1_CRESP'] == 'm') & (df['TwoBack1_RESP'] == 'm')].index)
    fa_count_3  = len(df[(df['procedure_Block_'] == 'threeproc')
                         & (df['ThreeBack1_CRESP'] == 'm') & (df['ThreeBack1_RESP'] == 'm')].index)
    
    # Calculate ratios of hits and false alarms
    data['hit_ratio_0'] = hit_count_0/M_COUNT
    data['hit_ratio_1'] = hit_count_1/M_COUNT
    data['hit_ratio_2'] = hit_count_2/M_COUNT
    data['hit_ratio_3'] = hit_count_3/M_COUNT
    data['fa_ratio_0'] = fa_count_0/X_COUNT
    data['fa_ratio_1'] = fa_count_1/X_COUNT
    data['fa_ratio_2'] = fa_count_2/X_COUNT
    data['fa_ratio_3'] = fa_count_3/X_COUNT
    
    return data


def extract_nback(proj, record_id, event_id, file_field, tmpdir):
    data = {}

    tab_file = download_file(proj, record_id, event_id, file_field, tmpdir)

    if tab_file is None:
        return None

    data = parse_nback(tab_file)

    return data


def load_nback(proj, record_id, event_id, file_field, tmpdir):
    nback_data = extract_nback(proj, record_id, event_id, file_field, tmpdir)

    if nback_data is None:
        logger.error('nback extract failed')
        return

    # Import the data
    data = {}
    data[proj.def_field] = record_id
    data['redcap_event_name'] = event_id
    data['nback_e90d1a_complete'] = '2'

    data['session'] = nback_data['Session']
    data['group'] = nback_data['Group']

    data['zero_back_fa'] = f'{nback_data["fa_ratio_0"]:.3f}'
    data['zero_back_hits'] = f'{nback_data["hit_ratio_0"]:.3f}'
    data['one_back_fa'] = f'{nback_data["fa_ratio_1"]:.3f}'
    data['one_back_hits'] = f'{nback_data["hit_ratio_1"]:.3f}'
    data['two_back_fa'] = f'{nback_data["fa_ratio_2"]:.3f}'
    data['two_back_hits'] = f'{nback_data["hit_ratio_2"]:.3f}'
    data['three_back_fa'] = f'{nback_data["fa_ratio_3"]:.3f}'
    data['three_back_hits'] = f'{nback_data["hit_ratio_3"]:.3f}'

    try:
        response = proj.import_records([data])
        assert 'count' in response
        logger.info(f'uploaded:{record_id}:{event_id}')
    except AssertionError as err:
        logger.error(f'uploaded:{record_id}:{event_id}')


logging.basicConfig(
    format='%(asctime)s - %(levelname)s:%(name)s:%(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

tmpdir = tempfile.mkdtemp()
logger.debug(f'tmpdir:{tmpdir}')

# Load the REDCap database
logger.info('INFO:loading REDCap...')
proj = redcap.Project(API_URL, API_KEY)
data = proj.export_records(events=REDCAP_EVENTS, fields=REDCAP_FIELDS)

for r in data:
    if not r['nback_tabfile']:
        logger.debug(f'no file:{r}')
        continue

    if not r['nback_tabfile'].endswith('.txt'):
        logger.debug(f'wrong format:{r}')
        continue

    if r['zero_back_fa']:
        logger.debug(f'already extracted:{r[proj.def_field]}:{r["nback_tabfile"]}')
        continue

    logger.info(f'loading:{r["nback_tabfile"]}')
    load_nback(proj, r[proj.def_field], r['redcap_event_name'], 'nback_tabfile', tmpdir)
       
logger.info('DONE!')
